# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/07_tens.ipynb.

# %% auto 0
__all__ = ['MemFmt', 'PresFmt', 'has_mps', 'devchain', 'try2dev', 'asdev', 'dtypechain', 'try2dtype', 'mpsdtype', 'to', 'ten2cpu',
           'ten2mps', 'ten2cuda', 'tenchain', 'try2ten', 'asten']

# %% ../nbs/07_tens.ipynb 6
from functools import wraps

# %% ../nbs/07_tens.ipynb 8
from typing import Optional, overload, TypeAlias, Union, ForwardRef

# %% ../nbs/07_tens.ipynb 9
try: from torch import memory_format as trcmemfmt, preserve_format as trcpresfmt
except ImportError: 
    class trchmemfmt(type): ...
    class trcpresfmt(trchmemfmt): ...

MemFmt = ForwardRef('torch.memory_format', is_class=True)
PresFmt = ForwardRef('torch.preserve_format', is_class=False)

# %% ../nbs/07_tens.ipynb 11
#| export


# %% ../nbs/07_tens.ipynb 13
try: import torch
except ImportError: ...

# %% ../nbs/07_tens.ipynb 15
try: import pytorch_lightning as pl
except ImportError: ...

# %% ../nbs/07_tens.ipynb 17
from atyp import IntQ, ListQ, Tensor, Device, AnyQ, DeviceQ, BoolQ, TensorQ, TorchType
from chck import isnone, isdict, notnone, istensor, istrc, isdevice, ismps, iscuda, isdtype
from calr import attrfunc, applyfns
from nlit import __DOC__, __NAME__, __MODULE__, __ANNOTATIONS__

# %% ../nbs/07_tens.ipynb 19
from .arrs import asarr, nparray, detach, clone, cpu, numpy, todense, tolist, toarray

# %% ../nbs/07_tens.ipynb 21
def has_mps() -> bool:
    '''Checks whether or not pytorch has mps availble (version) and was built with mps in mind.'''
    maybe_mps = torch.backends.mps.is_available()
    built_mps = torch.backends.mps.is_built()
    return maybe_mps and built_mps

# %% ../nbs/07_tens.ipynb 24
def devchain(x: DeviceQ, funcs, *args, **kwargs) -> Device:
    return applyfns(x, funcs, *args, check=isdevice, **kwargs)

# %% ../nbs/07_tens.ipynb 25
def try2dev(d: DeviceQ) -> Device:
    try: return d if isdevice(d) else torch.device(d)
    except: return d

# %% ../nbs/07_tens.ipynb 27
def asdev(d: DeviceQ, *args, **kwargs) -> Device:
    if isdevice(d): return d
    return devchain(d, (try2dev,), *args, **kwargs)

# %% ../nbs/07_tens.ipynb 30
def dtypechain(x: TorchType, funcs, *args, **kwargs) -> TorchType: 
    return applyfns(x, funcs, *args, check=isdtype, **kwargs)

# %% ../nbs/07_tens.ipynb 31
def try2dtype(a: Tensor, dtype: TorchType) -> Tensor:
    try: return a.to(dtype=dtype) if istensor(a) else torch.as_tensor(asarr(a), dtype=dtype)
    except: return a

# %% ../nbs/07_tens.ipynb 32
def mpsdtype(dtype: TorchType, device: DeviceQ = None, assume_on_mps: BoolQ = True) -> TorchType:
    '''Coerces `dtype` to `torch.float32` if `dtype` is `torch.float64` and `device` is `'mps'`.
    
    Parameters
    ----------    
    dtype : any
        dtype to check against
    
    device : Device, default=None
        the device of the tensor or model from which the `dtype` comes from. If provided
        will be used to detemine whether or not to make `torch.float64`, `torch.float32`
        only if the device is actually `'mps'`.

    assume_on_mps: bool, default=True
        whether or not to assume that the device of choice is `'mps'`. Setting this to
        `True` will result in `dtype` of `torch.float64` being converted to `torch.float32`
        to try and silently fix mps errors

    Notes
    -----
    - `float64` is not availble on `'mps'`, so they are coerced to `float32`
    - `assume_on_mps = True` and `has_mps() == True` are both needed as device might not 
    be provided.

    Returns
    -------
    dtype : any
        the dtype, corrected for mps if needed
    '''
    is_float64 = (dtype == torch.float64) or (dtype == 'float64') or (dtype == float)
    if notnone(device):
        if not isdevice(device): device = asdev(device)
        is_device_mps = device.type == 'mps'
        if is_device_mps:
            assume_on_mps = True

        elif device.type == 'cuda':
            assume_on_mps = False

    if has_mps() and assume_on_mps and is_float64: return torch.float32
    return dtype

# %% ../nbs/07_tens.ipynb 37
@overload
def to(
    x, other: Tensor, nonblocking: bool = False, copy: bool = False) -> Tensor: ...
@overload
def to(
    x, dtype, nonblocking: bool = False, copy: bool = False, 
    memory_format: MemFmt = trcpresfmt) -> Tensor: ...
@overload
def to(
    x, device: Device, dtype: TorchType = None, nonblocking: bool = False, copy: bool = False, 
    memory_format: MemFmt = trcpresfmt) -> Tensor: ...
@attrfunc('to', __all=True)
def to(x, *args, **kwargs) -> Tensor:
    '''Move tensor to device.'''

# %% ../nbs/07_tens.ipynb 40
@wraps(to, assigned = (__MODULE__, __DOC__, __ANNOTATIONS__))
def ten2cpu(t: Tensor, *args, **kwargs) -> Tensor:
    '''Given a tensor, ensures that it is on cpu.'''
    return to(t, device = asdev('cpu'))

@wraps(to, assigned = (__MODULE__, __DOC__, __ANNOTATIONS__))
def ten2mps(t: Tensor, *args, **kwargs) -> Tensor:
    '''Given a tensor, ensures that it is on mac silicon.    
    Additionally makes sure `tensor` is `torch.float32` if  `tensor.dtype` is `torch.float64`.'''
    
    if not istensor(t): return t
    dtype = mpsdtype(t.dtype, (device := asdev('mps')), assume_on_mps = False)
    return to(t, dtype = dtype, device = device)

@wraps(to, assigned = (__MODULE__, __DOC__, __ANNOTATIONS__))
def ten2cuda(t: Tensor, *args, **kwargs) -> Tensor:
    '''Given a tensor, ensures that it is on cuda.'''
    return to(t, device = asdev('cuda'))
    

# %% ../nbs/07_tens.ipynb 45
def tenchain(x: TensorQ, funcs, *args, **kwargs) -> Tensor: 
    return applyfns(x, funcs, *args, check=istensor, **kwargs)

# %% ../nbs/07_tens.ipynb 46
def try2ten(a: TensorQ, dtype: TorchType | None = None, device: DeviceQ = None) -> Tensor:
    try: return a if istensor(a) else torch.as_tensor(asarr(a), dtype=dtype, device=device)
    except: return a

# %% ../nbs/07_tens.ipynb 48
def asten(a, *args, **kwargs) -> Tensor:
    if istensor(a):
        return applyfns(a, (to, ), *args, **kwargs)

    chain = (asarr, try2move, try2ten, to, )
    a = tenchain(a, chain, *args, **kwargs)
    
    _mps, _cpu, _cuda = kwargs.get('mps', None), kwargs.get('cpu', None), kwargs.get('cuda', None)
    if _mps or ('mps' in args or asdev('mps') in args): chain += (ten2mps, )
    if _cpu or ('cpu' in args or asdev('cpu') in args): chain += (ten2cpu, )
    if _cuda or ('cuda' in args or asdev('cuda') in args): chain += (ten2cuda, )
    
    return applyfns(a, chain, *args, **kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_arrs.ipynb.

# %% auto 0
__all__ = ['chain', 'nparray', 'detach', 'clone', 'cpu', 'numpy', 'todense', 'tolist', 'toarray', 'arrchain', 'try2arr',
           'trc2arr', 'mtx2arr', 'spr2arr', 'dns2arr', 'ann2arr', 'mth2arr', 'asarr', 'as01']

# %% ../nbs/01_arrs.ipynb 6
#| export


# %% ../nbs/01_arrs.ipynb 8
from typing import Any, Union, TypeAlias, ParamSpec

# %% ../nbs/01_arrs.ipynb 10
try:
    import numpy as np
except ModuleNotFoundError:
    np = None

# %% ../nbs/01_arrs.ipynb 12
from atyp import StrQ, IntQ, NPArray, NPMatrix, Tensor, AnnData
from atyp import SPArray, SPMatrix, Index, Series, DataFrame

from chck import isad, isarr, ismtx, istens, issparse, istrc, isnone, notnone, isnpmatrix
from nlit import CPU, TOARRAY, TODENSE, TOLIST, DETACH, CLONE, NUMPY
from calr import attrfunc, applyfns
from achn import achn

# %% ../nbs/01_arrs.ipynb 14
#| export


# %% ../nbs/01_arrs.ipynb 17
class chain(achn): ...

@chain.able
def nparray(a): return np.array(a)

@chain.able
@attrfunc(DETACH)
def detach(x): '''try: x.detach() except: x'''

@chain.able
@attrfunc(CLONE)
def clone(x): '''try: x.clone() except: x'''

@chain.able
@attrfunc(CPU)
def cpu(x): '''try: x.cpu() except: x'''

@chain.able
@attrfunc(NUMPY)
def numpy(x): '''try: x.numpy() except: x'''

@chain.able
@attrfunc(TODENSE)
def todense(x): '''try: x.todense() except: x'''

@chain.able
@attrfunc(TOLIST)
def tolist(x): '''try: x.tolist() except: x'''

@chain.able
@attrfunc(TOARRAY)
def toarray(x): '''try: x.toarray() except: x'''

# %% ../nbs/01_arrs.ipynb 19
def arrchain(x, funcs, *args, **kwargs): 
    return applyfns(x, funcs, *args, check=isarr, **kwargs)

# %% ../nbs/01_arrs.ipynb 20
def try2arr(a) -> NPArray:
    try: return a if isarr(a) else np.array(a)
    except: return a

# %% ../nbs/01_arrs.ipynb 22
def trc2arr(a: Tensor) -> NPArray:
    '''from torch.tensor --> np.array e.g.
    `torch.tensor([...]).detach().clone().cpu().numpy()`'''
    if isarr(a) or not istrc(a): return a
    return arrchain(a, (detach, clone, cpu, numpy, ))

# %% ../nbs/01_arrs.ipynb 24
def mtx2arr(a: NPMatrix) -> NPArray:
    '''from np.matrix --> np.array e.g. 
    `np.array(np.matrix([...]).tolist())`'''
    if isarr(a) or not isnpmatrix(a): return a
    return arrchain(a, (tolist, nparray))

# %% ../nbs/01_arrs.ipynb 26
def spr2arr(a: Union[SPMatrix, SPArray]) -> NPArray:
    '''from scipy.sparse --> np.array e.g. 
    `scipy.sparse.csr_matrix([...]).toarray()`'''
    if isarr(a) or not issparse(a): return a
    return arrchain(a, (toarray, ))


# %% ../nbs/01_arrs.ipynb 28
def dns2arr(a: Union[SPMatrix, SPArray]) -> NPArray:
    '''from scipy.sparse --> np.array e.g. 
    `np.array(scipy.sparse.csr_matrix([...]).todense().tolist())`'''
    if isarr(a) or not issparse(a): return a
    return arrchain(a, (todense, tolist, nparray, ))

# %% ../nbs/01_arrs.ipynb 30
def ann2arr(a: AnnData, layer: StrQ = None) -> NPArray:
    if isarr(a) or not isad(a): return a
    arr = a.layers.get(layer, a.X)
    return arrchain(arr, (toarray, nparray))

# %% ../nbs/01_arrs.ipynb 32
def mth2arr(a, *args, **kwargs) -> NPArray:    
    if isarr(a): return a
    return arrchain(a, (
            spr2arr, # `scipy.sparse.csr_matrix([...]).toarray()`
            trc2arr, # `torch.tensor([...]).detach().clone().cpu().numpy()`
            dns2arr, # `np.array(scipy.sparse.csr_matrix([...]).todense().tolist())`
            mtx2arr, # `np.array(np.matrix([...]).tolist())`
            ann2arr, # `AnnData([...]).layers.get(layer, a.X).toarray()`
            try2arr, # `np.array`
        ), *args, **kwargs
    )

# %% ../nbs/01_arrs.ipynb 34
def asarr(a, *args, **kwargs) -> NPArray: return mth2arr(a, *args, **kwargs)

# %% ../nbs/01_arrs.ipynb 39
def as01(a: list, cutoff: float = 0.5) -> NPArray:
    '''binarize'''
    return (asarr(a) > cutoff).astype(int)



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/asto',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/asto',
                'lib_path': 'asto'},
  'syms': { 'asto.arng': { 'asto.arng._safe_idx': ('arng.html#_safe_idx', 'asto/arng.py'),
                           'asto.arng._safe_len': ('arng.html#_safe_len', 'asto/arng.py'),
                           'asto.arng.arange': ('arng.html#arange', 'asto/arng.py'),
                           'asto.arng.arngchain': ('arng.html#arngchain', 'asto/arng.py'),
                           'asto.arng.arr2arng': ('arng.html#arr2arng', 'asto/arng.py'),
                           'asto.arng.asarng': ('arng.html#asarng', 'asto/arng.py'),
                           'asto.arng.df2arng': ('arng.html#df2arng', 'asto/arng.py'),
                           'asto.arng.isarng': ('arng.html#isarng', 'asto/arng.py'),
                           'asto.arng.try2arng': ('arng.html#try2arng', 'asto/arng.py')},
            'asto.arrs': { 'asto.arrs.ann2arr': ('arrs.html#ann2arr', 'asto/arrs.py'),
                           'asto.arrs.arrchain': ('arrs.html#arrchain', 'asto/arrs.py'),
                           'asto.arrs.as01': ('arrs.html#as01', 'asto/arrs.py'),
                           'asto.arrs.asarr': ('arrs.html#asarr', 'asto/arrs.py'),
                           'asto.arrs.chain': ('arrs.html#chain', 'asto/arrs.py'),
                           'asto.arrs.clone': ('arrs.html#clone', 'asto/arrs.py'),
                           'asto.arrs.cpu': ('arrs.html#cpu', 'asto/arrs.py'),
                           'asto.arrs.detach': ('arrs.html#detach', 'asto/arrs.py'),
                           'asto.arrs.dns2arr': ('arrs.html#dns2arr', 'asto/arrs.py'),
                           'asto.arrs.mth2arr': ('arrs.html#mth2arr', 'asto/arrs.py'),
                           'asto.arrs.mtx2arr': ('arrs.html#mtx2arr', 'asto/arrs.py'),
                           'asto.arrs.nparray': ('arrs.html#nparray', 'asto/arrs.py'),
                           'asto.arrs.numpy': ('arrs.html#numpy', 'asto/arrs.py'),
                           'asto.arrs.spr2arr': ('arrs.html#spr2arr', 'asto/arrs.py'),
                           'asto.arrs.toarray': ('arrs.html#toarray', 'asto/arrs.py'),
                           'asto.arrs.todense': ('arrs.html#todense', 'asto/arrs.py'),
                           'asto.arrs.tolist': ('arrs.html#tolist', 'asto/arrs.py'),
                           'asto.arrs.trc2arr': ('arrs.html#trc2arr', 'asto/arrs.py'),
                           'asto.arrs.try2arr': ('arrs.html#try2arr', 'asto/arrs.py')},
            'asto.cast': {'asto.cast.cast': ('cast.html#cast', 'asto/cast.py'), 'asto.cast.caster': ('cast.html#caster', 'asto/cast.py')},
            'asto.dict': {},
            'asto.nums': { 'asto.nums.ascomplex': ('nums.html#ascomplex', 'asto/nums.py'),
                           'asto.nums.asfloat': ('nums.html#asfloat', 'asto/nums.py'),
                           'asto.nums.asint': ('nums.html#asint', 'asto/nums.py'),
                           'asto.nums.asnum': ('nums.html#asnum', 'asto/nums.py')},
            'asto.sets': { 'asto.sets.any2set': ('sets.html#any2set', 'asto/sets.py'),
                           'asto.sets.asset': ('sets.html#asset', 'asto/sets.py'),
                           'asto.sets.setchain': ('sets.html#setchain', 'asto/sets.py'),
                           'asto.sets.try2set': ('sets.html#try2set', 'asto/sets.py')},
            'asto.slcs': { 'asto.slcs.asslc': ('slcs.html#asslc', 'asto/slcs.py'),
                           'asto.slcs.safeslice': ('slcs.html#safeslice', 'asto/slcs.py'),
                           'asto.slcs.slcchain': ('slcs.html#slcchain', 'asto/slcs.py'),
                           'asto.slcs.slice2list': ('slcs.html#slice2list', 'asto/slcs.py'),
                           'asto.slcs.try2slc': ('slcs.html#try2slc', 'asto/slcs.py')},
            'asto.tens': { 'asto.tens.asdev': ('tens.html#asdev', 'asto/tens.py'),
                           'asto.tens.asten': ('tens.html#asten', 'asto/tens.py'),
                           'asto.tens.devchain': ('tens.html#devchain', 'asto/tens.py'),
                           'asto.tens.dtypechain': ('tens.html#dtypechain', 'asto/tens.py'),
                           'asto.tens.has_mps': ('tens.html#has_mps', 'asto/tens.py'),
                           'asto.tens.mpsdtype': ('tens.html#mpsdtype', 'asto/tens.py'),
                           'asto.tens.ten2cpu': ('tens.html#ten2cpu', 'asto/tens.py'),
                           'asto.tens.ten2cuda': ('tens.html#ten2cuda', 'asto/tens.py'),
                           'asto.tens.ten2mps': ('tens.html#ten2mps', 'asto/tens.py'),
                           'asto.tens.tenchain': ('tens.html#tenchain', 'asto/tens.py'),
                           'asto.tens.to': ('tens.html#to', 'asto/tens.py'),
                           'asto.tens.try2dev': ('tens.html#try2dev', 'asto/tens.py'),
                           'asto.tens.try2dtype': ('tens.html#try2dtype', 'asto/tens.py'),
                           'asto.tens.try2ten': ('tens.html#try2ten', 'asto/tens.py')},
            'asto.time': { 'asto.time.day2hours': ('time.html#day2hours', 'asto/time.py'),
                           'asto.time.day2hrs': ('time.html#day2hrs', 'asto/time.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_cast.ipynb.

# %% auto 0
__all__ = ['T', 'cast', 'caster']

# %% ../nbs/00_cast.ipynb 6
from functools import wraps

# %% ../nbs/00_cast.ipynb 8
from typing import Any, Union, TypeVar
T = TypeVar('T')

# %% ../nbs/00_cast.ipynb 10
#| export


# %% ../nbs/00_cast.ipynb 12
#| export


# %% ../nbs/00_cast.ipynb 14
#| export


# %% ../nbs/00_cast.ipynb 16
def cast(x, dtype: T, retnone: bool = False) -> Union[T, Any, None]:
    '''Tries to cast `x` to `dtype`. If `retnone` is `True`, returns `None` if the cast fails, otherwise returns `x`.'''
    try: return dtype(x)
    except: return None if retnone else x


def caster(dtype: T, retnone: bool = False):
    def decorator(func):
        @wraps(func)
        def wrapper(x, retnone: bool = retnone) -> Union[T, Any, None]:
            return cast(x, dtype, retnone)
        return wrapper
    return decorator



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/08_sets.ipynb.

# %% auto 0
__all__ = ['try2set', 'any2set', 'setchain', 'asset']

# %% ../nbs/08_sets.ipynb 6
#| export


# %% ../nbs/08_sets.ipynb 8
from typing import Optional

# %% ../nbs/08_sets.ipynb 10
#| export


# %% ../nbs/08_sets.ipynb 12
from atyp import IntQ, ListQ
from calr import applyfns
from chck import isiter, islist, istuple, isset, isslice, isiter, isiternotstr

# %% ../nbs/08_sets.ipynb 14
#| export


# %% ../nbs/08_sets.ipynb 16
def try2set(a) -> set:
    if isset(a): return a
    try: return set([a]) if not isiternotstr(a) else set(a)
    except: return a

# %% ../nbs/08_sets.ipynb 17
def any2set(obj) -> set:
    '''Tries to convert any object to a set. If the object is a string, it will be converted to a set of length 1.'''    
    if isinstance(obj, str): obj = set((obj, ))
    elif not isiter(obj): obj = set((obj, )) # not a string and not iterable

    elif any(c(obj) for c in (isiter, islist, istuple)) and not isset(obj):    
        obj = set(obj)
    elif hasattr(obj, '__hash__'): obj = {obj, }
    else:
        try: obj = set(obj)
        except TypeError: obj = set()
    return obj if isset(obj) else set()

# %% ../nbs/08_sets.ipynb 18
def setchain(x, funcs, *args, **kwargs): 
    return applyfns(x, funcs, *args, check=isset, **kwargs)

# %% ../nbs/08_sets.ipynb 20
def asset(a, *args, **kwargs) -> slice:
    if isset(a): return a
    chain = (any2set, try2set, )
    return setchain(a, chain, *args, **kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_dict.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/04_dict.ipynb 6
#| export


# %% ../nbs/04_dict.ipynb 8
from typing import Any, Union, TypeAlias, ParamSpec

# %% ../nbs/04_dict.ipynb 10
#| export


# %% ../nbs/04_dict.ipynb 12
#| export


# %% ../nbs/04_dict.ipynb 14
#| export


# %% ../nbs/04_dict.ipynb 16
#| export



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_time.ipynb.

# %% auto 0
__all__ = ['day2hrs', 'day2hours']

# %% ../nbs/02_time.ipynb 6
from datetime import date, datetime

# %% ../nbs/02_time.ipynb 8
from typing import Any, Union, TypeAlias, ParamSpec

# %% ../nbs/02_time.ipynb 10
#| export


# %% ../nbs/02_time.ipynb 12
#| export


# %% ../nbs/02_time.ipynb 14
#| export


# %% ../nbs/02_time.ipynb 16
def day2hrs(day: int) -> int:
    '''Convert days to hours
    
    See Also
    --------
    day2hours : Alias of `day2hrs`
    '''
    return int(day * 24)

def day2hours(day: int) -> int:
    '''Convert days to hours
    
    See Also
    --------
    day2hrs : Alias of `day2hours`
    '''
    return day2hrs(day)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_nums.ipynb.

# %% auto 0
__all__ = ['asint', 'asfloat', 'ascomplex', 'asnum']

# %% ../nbs/03_nums.ipynb 6
#| export


# %% ../nbs/03_nums.ipynb 8
from typing import Any, Union, TypeAlias, ParamSpec

# %% ../nbs/03_nums.ipynb 10
#| export


# %% ../nbs/03_nums.ipynb 12
from atyp import NumberQ
from calr import applyfns
from chck import isnum

# %% ../nbs/03_nums.ipynb 14
from .cast import caster

# %% ../nbs/03_nums.ipynb 16
@caster(int, retnone=False)
def asint(x, retnone: bool = False) -> Union[int, None]:
    'Converts `x` to an `int` if possible, otherwise returns `None`.'

@caster(float, retnone=False)
def asfloat(x, retnone: bool = False) -> Union[float, None]:
    'Converts `x` to an `int` if possible, otherwise returns `None`.'

@caster(complex, retnone=False)
def ascomplex(x, retnone: bool = False) -> Union[float, None]:
    'Converts `x` to an `int` if possible, otherwise returns `None`.'

# %% ../nbs/03_nums.ipynb 17
def asnum(x, retnone: bool = False) -> NumberQ:
    return applyfns(x, (asint, asfloat, ascomplex, ), check=isnum, retnone=retnone)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_slcs.ipynb.

# %% auto 0
__all__ = ['safeslice', 'slice2list', 'try2slc', 'slcchain', 'asslc']

# %% ../nbs/06_slcs.ipynb 6
#| export


# %% ../nbs/06_slcs.ipynb 8
from typing import Optional

# %% ../nbs/06_slcs.ipynb 10
#| export


# %% ../nbs/06_slcs.ipynb 12
from atyp import IntQ, ListQ
from chck import isnone, isdict, notnone, isslice, isiter
from calr import applyfns

# %% ../nbs/06_slcs.ipynb 14
#| export


# %% ../nbs/06_slcs.ipynb 16
def safeslice(s: slice) -> slice:
    '''Returns a slice with default values instead None values from the input slice.
    e.g. slice(IntQ, IntQ, IntQ) -> slice(0, 1, 1)
    '''
    if not isslice(s): return s
    start = s.start if notnone(s.start) else 0
    stop  = s.stop  if notnone(s.stop)  else min(0, start, max(1, start))
    step  = s.step  if notnone(s.step)  else 1
    return slice(start, stop, step)

def slice2list(s: slice) -> tuple[IntQ, int, IntQ]:
    '''Returns a list of integers from a slice.'''    
    s = safeslice(s)
    try:    return list(range(s.start, s.stop, s.step))
    except: return list()

def try2slc(a) -> slice:
    if isslice(a): return a
    try: return slice(*list((a, )if not isiter(a) else a))    
    except: return a

# %% ../nbs/06_slcs.ipynb 17
def slcchain(x, funcs, *args, **kwargs): 
    return applyfns(x, funcs, *args, check=isslice, **kwargs)

# %% ../nbs/06_slcs.ipynb 19
def asslc(a, *args, **kwargs) -> slice:
    if isslice(a): return a
    chain = (safeslice, try2slc, )
    slc = slcchain(a, chain, *args, **kwargs)
    if kwargs.pop('__nones_ok', False): return slc
    return safeslice(slc)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_arng.ipynb.

# %% auto 0
__all__ = ['isarng', 'arngchain', 'try2arng', 'arr2arng', 'df2arng', 'asarng', 'arange']

# %% ../nbs/05_arng.ipynb 6
#| export


# %% ../nbs/05_arng.ipynb 8
from typing import TypeGuard

# %% ../nbs/05_arng.ipynb 10
try: import numpy as np
except ModuleNotFoundError: np = None

# %% ../nbs/05_arng.ipynb 12
from atyp import NPArray, IndexLike, DataFrame
from chck import isarr
from calr import applyfns
# from nlit import INDEX

# %% ../nbs/05_arng.ipynb 14
from .arrs import arrchain, try2arr, asarr

# %% ../nbs/05_arng.ipynb 17
def _safe_idx(x): 
    '''try: x.index except: x'''
    try: return x.index
    except: return x

def _safe_len(x):
    '''try: len(x) except: 1'''
    try: return len(x)
    except: return 1

def isarng(a) -> TypeGuard[NPArray]:
    return isarr(a) and np.all(np.ediff1d(a) == 1)

def arngchain(x, funcs, *args, **kwargs): 
    return applyfns(x, funcs, *args, check=isarng, **kwargs)

def try2arng(a) -> NPArray:
    if isarng(a): return a
    try:
        arange = np.arange(_safe_len(a))
        return arange
    except: return a

def arr2arng(a: IndexLike) -> NPArray:
    if isarng(a): return a
    arange = np.arange(len(asarr(a)))
    return arange

def df2arng(df: DataFrame) -> NPArray:
    if isarng(df): return df
    return arr2arng(_safe_idx(df))

# %% ../nbs/05_arng.ipynb 19
def asarng(a, *args, **kwargs) -> NPArray:    
    if isarng(a): return a
    return arngchain(a, (
            df2arng, 
            arr2arng,
            try2arng,            
        ), *args, **kwargs
    )

# %% ../nbs/05_arng.ipynb 20
def arange(a, *args, **kwargs) -> NPArray: 
    return asarng(a, *args, **kwargs)



__version__ = "0.0.16"
# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00__init__.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/00__init__.ipynb 3
import itertools
from importlib import import_module
from types import ModuleType

def loadmod(mod: ModuleType):
    if hasattr(mod, '__all__'): return mod.__all__
    if hasattr(mod, '__module__'): return loadmod(import_module(mod.__module__))
    raise AttributeError(f"module {mod} has no attribute '__all__'")

# %% ../nbs/00__init__.ipynb 4
from asto import (
    cast, arrs, time, nums, dict, arng, slcs, tens, sets
)

from .cast import *
from .arrs import *
from .time import *
from .nums import *
from .dict import *
from .arng import *
from .slcs import *
from .tens import *
from .sets import *

# %% ../nbs/00__init__.ipynb 5
__all__ = sorted(set(itertools.chain(*[
    loadmod(mod) for mod in (cast, arrs, time, nums, dict, arng, slcs, tens, sets)
])))
